; 68000 music player for Saturn

; assemble with vasmm68k_mot

; set the friggin memory size in $400 from the SH2 side before writing
; anything to sound RAM or shit will be all fucked up!

; SCSP registers
; $100000-$1003F8 individual slot control registers (12 16-bit words each)
;
;  $00  bit 12 = key on execute      \ writing a 1 here performs key-on or
;       bit 11 = 0=key-off,1=key-on   \ key-off for every slot simultaneously!
;       bit 10 = reverse source sign bit
;       bit 9 = reverse source data bits 
;       bits 8-7 = sound source control: 0=PCM, 1=noise, 2=silence
;       bits 6-5 = loop control: 0=no loop, 1=normal loop
;                  2=reverse loop (plays backward from loopend to loopstart)
;                  3="alternative" loop (plays fwd,back,fwd,back,etc.)
;       bit 4 = PCM sample size: 0=16bit, 1=8bit (both '2's complement)
;       bits 3-0 = high 4 bits of start address
;
;  $02  low 16 bits of start address (16bit PCM must be word-aligned)
;
;  $04  loop start address \
;                           \  measured in number of samples from start addr
;  $06  loop end address     \
;
;  $08  bits 15-11 = decay 2            bits 10-6 = decay 1
;       bit 5 = EG hold mode (volume goes to max immediately after key on?)
;       bits 4-0 = attack rate
;
;  $0A  bit 14 = loop start link (waveform begins loop upon reaching decay 1)
;       bits 13-10 = key scaling        bits 9-5 = decay level (0=max volume)
;       bits 4-0 = release rate
;
;  $0C  bit 9 = stack write inhibit
;       bit 8 = sound direct mode: 1=skip LFO, envelope generator
;       bits 7-0 = total level (0=max volume)
;
;  $0E  bits 15-12 = modulation level
;       bits 11-6 = modulation input X select
;       bits 5-0 = modulation input Y select
;       input selects are relative to the number of the carrier slot
;       slot 1, select $21 = input from current sample of slot 2
;       slot 1, select $01 = input from old sample of slot 2
;       (if low 5 bits are $1C or greater, than MSB function is reversed)
;       (because of the timing/order in which slot data is written to stack)
;       slot 1, select $1F (not $3F) = input from current sample of slot 0
;
;  $10  bits 14-11 = octave (signed!)   bits 9-0 = frequency number
;       (octave=0,fn=0 tone matches source sampled at 44100hz,
;        octave=-1 is half frequency, octave=1 is double)
;
;  $12  bit 15 = LFO reset (1 stops oscillation)
;       bits 14-10 = LFO frequency (0=.17hz, $1F=172.3hz)
;       bits 9-8 = LFO FM waveform      bits 7-5 = FM amplitude
;       bits 4-3 = LFO AM waveform      bits 2-0 = AM amplitude
;       waveform 0=saw, 1=square, 2=triangle, 3=noise
;
;  $14  bits 6-3 = input select (used for sending output to DSP mix stack)
;       bits 2-0 = input mix level 0=off, 1=1/64 volume, 7=full volume
;
;  $16  bits 15-13 = direct data level  bits 12-8 = direct data panning
;       bits 7-5 = effect data level    bits 4-0 = effect data panning
;       panning $00=$10=center, $0F=all right, $1F=all left
;
; Global control registers
;
; $100400  bit 9 = mem size 0=1Mbit DRAM, 1=4Mbit DRAM (normal setting)
;          bit 8 = DAC size 0=16bit (normal setting), 1=18bit
;          bits 3-0 = master volume
;
; $100402  bits 8-7 = ring buffer length       \  (DSP related)
;          bits 6-0 = ring buffer lead address  \
;                     (forms bits 19-13 of the address)
;
; $100404  bit 12 = output FIFO full     bit 11 = output FIFO empty
;          bit 10 = input FIFO overflow
;          bit 9 = input FIFO full       bit 8 = input FIFO empty
;          bits 7-0 = MIDI input buffer
;
; $100406  bits 7-0 = MIDI output buffer
;
; $100408  bits 15-11 = number of slot to monitor
;          bits 10-7 = sample data position (times 4096) to monitor
;
; $100412  DMA start address
;
; $100414  bits 15-12 = high bits of DMA start address
;          bits 11-1 = DMA register address (addresses must be even)
;
; $100416  bit 14 = gate clear
;          bit 13 = transfer direction
;          bit 12 = start transfer
;          bits 11-1 = transfer length
;
; $100418  bits 10-8 = timer A pre-scaler control
;          bits 7-0 = timer A counter
; $10041A  bits 10-8 = timer B pre-scaler control
;          bits 7-0 = timer B counter
; $10041C  bits 10-8 = timer C pre-scaler control
;          bits 7-0 = timer C counter
;            pre-scaler control 0=counter increment every sample (44KHz)
;                               1=every 2nd sample
;                               7=every 128th sample, etc.
;
; $10041E  bits 10-0 = sound CPU interrupt enable bits
; $100420  bits 10-0 = sound CPU interrupt status bits
;                        (bit5 can be written to trigger an interrupt)
; $100422  bits 10-0 = sound CPU interrupt reset bits
;
; $100424  bits 7-0 = sound CPU intterrupt level bit0
; $100426  bits 7-0 = sound CPU intterrupt level bit1
; $100428  bits 7-0 = sound CPU intterrupt level bit2
;               corresponding bits from each word are combined to specify
;               the sound CPU interrupt level for each interrupt source
;               (top four ints share the value at bit 7)
;
; $10042A  bits 10-0 = main CPU interrupt enable bits
; $10042C  bits 10-0 = main CPU interrupt status bits
;                        (bit5 can be written to trigger an interrupt)
; $10042E  bits 10-0 = main CPU interrupt reset bits
;
; interrupts    bit 10=every sample             bit 9=MIDI output    
;               bit 8=timer C                   bit 7=timer B
;               bit 6=timer A                   bit 5=CPU
;               bit 4=DMA transfer done         bit 3=MIDI input
;               bit 2=external INT2N            bit 1=external INT1N
;               bit 0=external INT0N
;
; $100600-3F  Sound stack A
; $100640-7F  Sound stack B
;
; $100700-EE3  DSP area
;

        machine 68000

        org 0

; vector area
        dc.l $7FE,main
        dc.l nullint,nullint,nullint,nullint,nullint,nullint
        dc.l nullint,nullint,nullint

        org $800

nullint:
        rts

main:

        move.w #$2700,SR       ; set supervisor mode, disable interrupts

        lea $400,a6          ; base address for variables

        ; write a value to notify SH2 that we are running
        move.w #$ABCD,d0
        move.w d0,(a6)

        jsr initslots

        ; setup instruments
        lea scspinitregs,a0
        move.l #$100000,d0
a100:
        move.w (a0)+,d0
        bmi a101
        move.l d0,a1
        move.w (a0)+,d0
        move.w d0,(a1)
        bra a100
a101:

        ; init music player
        lea track1,a0
        move.l a0,$10(a6)
        lea track2,a0
        move.l a0,$18(a6)
        lea track3,a0
        move.l a0,$20(a6)
        lea track4,a0
        move.l a0,$28(a6)

        moveq #0,d0
        move.w d0,$14(a6)
        move.w d0,$1C(a6)
        move.w d0,$24(a6)
        move.w d0,$2C(a6)

; $410 - pointer to track 1 data
; $414 - track 1 timer (counts down every vblank)
;
; $418 - pointer to track 2 data
; $41C - track 2 timer (...)
;
; $420 - pointer to track 3 data
; $424 - track 3 timer (...)
;
; $428 - pointer to track 4 data
; $42C - track 4 timer (...)

        lea $100000,a5

        ; play stuff
a113:
        jsr waitv
        lea $410,a4
        moveq #3,d7
a112:
        move.l (a4),a3
        move.w 4(a4),d0
        bne a114
        ; load some more crap from track data
a115:
        move.w (a3)+,d1
        bmi a116
        ; load new timer value
;        addq.w #1,d1
        move.w d1,d0
        bra a114
a116:
        ; execute command
        ; first calculate slot address
        move.w d1,d2
        asr.w #3,d2
        and.w #$3E0,d2
        cmp.b #0,d1
        bne a117
        ; key ON, with key off immediately preceding
        move.b #$10,0(a5,d2)
        jsr wastetime
        move.b #$18,0(a5,d2)
        bra a115
a117:
        cmp.b #1,d1
        bne a118
        ; JUMP to elsewhere in track data
        move.l (a3)+,d3
        move.l d3,a3
        bra a115
a118:
        cmp.b #2,d1
        bne a119
        ; key ON, for a 4-op instrument
        move.b #0,0(a5,d2)
        move.b #0,$20(a5,d2)
        move.b #0,$40(a5,d2)
        move.b #$10,$60(a5,d2)
        jsr wastetime
        move.b #8,0(a5,d2)
        move.b #8,$20(a5,d2)
        move.b #8,$40(a5,d2)
        move.b #$18,$60(a5,d2)
        bra a115
a119:
        cmp.b #3,d1
        bne a120
        ; key OFF, for a 4-op instrument
        move.b #0,0(a5,d2)
        move.b #0,$20(a5,d2)
        move.b #0,$40(a5,d2)
        move.b #$10,$60(a5,d2)
        bra a115
a120:
        cmp.b #4,d1
        bne a121
        ; set FREQUENCY, for a 4-op instrument with 2-4-1-1 ratios
        move.w (a3)+,d3
        move.w d3,$50(a5,d2)
        move.w d3,$70(a5,d2)
        add.w #$800,d3
        move.w d3,$10(a5,d2)
        add.w #$800,d3
        move.w d3,$30(a5,d2)
        bra a115
a121:
        cmp.b #5,d1
        bne a122
        ; set FREQUENCY, for a 4-op instrument with .5-2-4-1 ratios
        move.w (a3)+,d3
        move.w d3,$70(a5,d2)
        add.w #$800,d3
        move.w d3,$30(a5,d2)
        add.w #$800,d3
        move.w d3,$50(a5,d2)
        sub.w #$1800,d3
        move.w d3,$10(a5,d2)
        bra a115
a122:

a114:
        move.l a3,(a4)+
        ; decrement timer and advance to next track
        subq.w #1,d0
        move.w d0,(a4)
        add.w #4,a4
        dbra d7,a112
        bra a113




wastetime:
        move.l d0,-(a7)
        moveq #100,d0
timewaste:
        nop
        dbra d0,timewaste
        move.l (a7)+,d0
        rts

waitv:
        move.w $406,d7
waitv2:
        cmp.w $406,d7
        beq waitv2
        rts

initslots:
        lea $100000,a0
        moveq #32,d1
a107:
        move.w #$0022,(a0)+
        move.w #$C800,(a0)+
        move.w #0,(a0)+
        move.w #$0400,(a0)+
        move.w #31,(a0)+
        move.w #31,(a0)+

;        move.w #$0100,(a0)+
        move.w #0,(a0)+

        move.w #0,(a0)+
        move.w #0,(a0)+
        move.w #0,(a0)+
;        move.w #7,(a0)+
;        move.w #$E0E0,(a0)+
        move.w #0,(a0)+
        move.w #$E000,(a0)+

        add.w #8,a0
        subq.w #1,d1
        bne a107
        rts


scspinitregs:
;        dc.w $0400,$020F
;        dc.w $0402,$0118
        dc.w $0402,0
        dc.w $041E,0
        dc.w $042A,0

        ; slot 0
        dc.w $0000,$0001
        dc.w $0002,0
        dc.w $0006,$1FFE
        dc.w $000C,$0100
        dc.w $0010,$7000

        ; slot 1
        dc.w $0020,$0001
        dc.w $0022,$4000
        dc.w $0026,$1FFE
        dc.w $002C,$0100
        dc.w $0030,$7000

        ; slot 2
        dc.w $0040,$0001
        dc.w $0042,$8000
        dc.w $0046,$1FFE
        dc.w $004C,$0100
        dc.w $0050,$7000

        ; slot 3
        dc.w $0060,$0001
        dc.w $0062,$C000
        dc.w $0066,$1FFE
        dc.w $006C,$0100
        dc.w $0070,$7000

        ; slot 4
        dc.w $0080,$0002
        dc.w $0082,0
        dc.w $0086,$1FFE
        dc.w $008C,$0100
        dc.w $0090,$7000

        ; slot 5
        dc.w $00A0,$0002
        dc.w $00A2,$4000
        dc.w $00A6,$1FFE
        dc.w $00AC,$0100
        dc.w $00B0,$7000

        ; slot 6
        dc.w $00C0,$0002
        dc.w $00C2,$8000
        dc.w $00C6,$1FFE
        dc.w $00CC,$0100
        dc.w $00D0,$7000

        ; slot 7
        dc.w $00E0,$0002
        dc.w $00E2,$C000
        dc.w $00E6,$1FFE
        dc.w $00EC,$0100
        dc.w $00F0,$7000

; a 4-op FM melody instrument

        ; slot 8          (2)
        dc.w $0108,$2B9F
        dc.w $010A,$0004
        dc.w $010C,$0040
        dc.w $0110,$2000
;        dc.w $0112,$1A73
        dc.w $0116,0

        ; slot 9           (4)
        dc.w $0128,$2CDF
        dc.w $012A,$00CA
        dc.w $012C,$0030
        dc.w $0130,$2800
;        dc.w $0132,$1A73
        dc.w $0136,0

        ; slot 10          (1)
        dc.w $0148,$1C5F
        dc.w $014A,$000A
        dc.w $014C,$0020
        dc.w $014E,$D79E
        dc.w $0150,$1800
        dc.w $0152,$5230

        ; slot 11        (3)
        dc.w $0168,$2458
        dc.w $016A,$000A
        dc.w $016C,$0020
        dc.w $016E,$E79E
        dc.w $0170,$1800
        dc.w $0172,$5230

; a 4-op FM bass instrument

        ; slot 12        (3)
        dc.w $0188,$2458
        dc.w $018A,$000A
        dc.w $018C,$0030
        dc.w $0190,$0000
;        dc.w $0192,$5230

        ; slot 13          (2)
        dc.w $01A8,$2B9F
        dc.w $01AA,$0004
        dc.w $01AC,$0040
        dc.w $01AE,$E7DF
        dc.w $01B0,$1000
;        dc.w $01B2,$1A73
        dc.w $01B6,0

        ; slot 14           (4)
        dc.w $01C8,$2CDF
        dc.w $01CA,$00CA
        dc.w $01CC,$0030
        dc.w $01D0,$1800
;        dc.w $01D2,$1A73
        dc.w $01D6,0

        ; slot 15          (1)
        dc.w $01E8,$1C5F
        dc.w $01EA,$000A
        dc.w $01EC,$0020
        dc.w $01EE,$D79F
        dc.w $01F0,$0800
        dc.w $01F2,$5230

        dc.w $8000

; 0=key on 1-op
; 1=jump
; 2=key on 4-op
; 3=key off 4-op
; 4=set freq 4-op (pfm instrument)
; 5=set freq 4-op (bass instrument)

track1:
        dc.w 128,$8100,32,$8500,32,$8000,64,$8001
        dc.l track1
track2:
        dc.w $8804,$1A12,$8802,256
        dc.w $8804,$1AD0,$8802,256
        dc.w $8804,$19BA,$8802,256
        dc.w $8804,$1A12,$8802,256
        dc.w $8804,$191A,$8802,256
        dc.w $8804,$188C,$8802,256
        dc.w $8804,$19BA,$8802,256
        dc.w $8804,$1A12,$8802,256
        dc.w $8001
        dc.l track2
track3:
        dc.w $8C05,$03A5,$8C02,32       ; E
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w $8C05,$091A,$8C02,32       ; A
        dc.w $8C05,$088C,$8C02,32       ; G
        dc.w 128
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w $8C05,$088C,$8C02,32       ; G
        dc.w $8C05,$03A5,$8C02,32       ; E
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w 128
        dc.w $8C05,$02D0,$8C02,32       ; D
        dc.w $8C05,$0212,$8C02,32       ; C (low)
        dc.w $8C05,$02D0,$8C02,32       ; D
        dc.w $8C05,$03A5,$8C02,32       ; E
        dc.w 128
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w $8C05,$088C,$8C02,32       ; G
        dc.w $8C05,$0A12,$8C02,32       ; C (high)
        dc.w $8C05,$09BA,$8C02,32       ; B
        dc.w 128
        dc.w $8C05,$091A,$8C02,32       ; A
        dc.w $8C05,$03A5,$8C02,32       ; E
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w $8C05,$088C,$8C02,32       ; G
        dc.w 128
        dc.w $8C05,$088C,$8C02,32       ; G
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w $8C05,$03A5,$8C02,32       ; E
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w 128
        dc.w $8C05,$091A,$8C02,32       ; A
        dc.w $8C05,$088C,$8C02,32       ; G
        dc.w $8C05,$03A5,$8C02,32       ; E
        dc.w $8C05,$080D,$8C02,32       ; F
        dc.w 128
        dc.w $8C05,$088C,$8C02,32       ; G
        dc.w $8C05,$091A,$8C02,32       ; A
        dc.w $8C05,$09BA,$8C02,32       ; B
        dc.w $8C05,$0A12,$8C02,32       ; C (high)

        dc.w 128,$8001
        dc.l track3
track4:
        dc.w 128,$8001
        dc.l track4
