; PS1 credits dungeon maze demo by DamageX

; assemble with SUPERAS

        setpc $06004000

; program RAM usage
;
; 60DF000 number of verteces
; 60DF004 number of polygons
; 60DF008 misc (x position to draw walls)
; 60DF00C axz
;  60DF00E angular velocity
; 60DF010 address of command table being built
; 60DF014 Z position in tunnel section    \   -128 to 127
;  60DF016 X position in tunnel section    \   (both 0 when not moving)
; 60DF018 misc (depth to draw walls)
; 60DF01C vblank frame counter
; 60DF020 controller status (32 bits)
;               bits 27-24 - Rt X Y Z (works on Saturn, not Yabause)
;               bits 19-16 - Start A C B
;               bits 11-8  - R L D U
; 60DF024 old controller status
; 60DF028 button press detection
; 60DF02C Y/Z map position
;  60DF02E X map position
; 60DF030 movement direction (0=none, else bits 3-0 = R L D U)
; 60DF034 pointer to sometext
; 60DF038 pointer to autoplay script
; 60DF03C script timer
;
; 60DF040-B sin, cos, -sin for viewing angle
;
; 60E0000 untransformed vertex data
; 60EC000 polygon texture assignments
; 60F0000 transformed vertex data
; 60F8000 polygon list


; SMPC registers (byte accesses only)
; $20100001-D  W (odd addresses only) input registers 0-6
; $2010001F    W command register
; $20100021-5F R (odd addresses only) output registers 0-31 
; $20100061    R status register
; $20100063    R/W status flag (write $01 before issuing command)
;                               (bit 0 is cleared when SMPC is done)
; $20100075    R/W PDR1 \
; $20100077    R/W PDR2  \ 7-bit parallel ports (controller ports)
;       first byte - R L D U Start A C B  \
;       second byte - Rt X Y Z Lt x x x    \ (for SMPC polling mode)
; $20100079    W DDR1   \
; $2010007B    W DDR2    \ 7-bit data direction registers (1=output)
; $2010007D    W bit 0 = IOSEL1, bit 1 = IOSEL2
;               set IOSEL for SH2 direct access, else polled by SMPC
; $2010007F    W bit 0 = EXLE1, bit 1 = EXLE2
;               EXLE = external latch enable (used for light guns)
; SMPC commands
; (OREG31 will contain the command code after command issue)
; $00 master sh2 on
; $02 slave sh2 on 
; $03 slave sh2 off
; $06 sound on
; $07 sound off
; $08 CD on
; $09 CD off
; $0D system reset
; $0E change to 28.6MHz
; $0F change to 26.8MHz
; $10 INTBACK (used to read time, peripheral data, and region codes)
; $16 RTC set time
; $17 SMPC memory setting
; $18 NMI request (?)
; $19 reset enable (?)
; $1A reset disable (?)



        ; change system clock to 28.6MHz

        imov.l $6000320,r1
        mov.l @r1,r1
        jsr @r1
        nop

        ; setup vblank interrupt routine

; 25FE00A0 - SCU interrupt control register 
; setting a bit masks the interrupt
; bit 0 is VBlank (vector $40)

        imov.l $6000100,r1
        mova (vbint,PC),r0
        mov.l r0,@r1

        imov.l $25FE00A0,r2
        imov.l $FFFFBFFE,r4
        mov.l r4,@r2

        imov.l $E0,r1
        ldc r1,SR

        bra main
        nop

        align.l

vbint:
        mov.l r0,@-r15
        mov.l r1,@-r15
        mov.l r2,@-r15
        mov.l r4,@-r15

        ; increment frame counter
        imov.l $60DF01C,r2
        mov.l @r2,r4
        add #1,r4
        mov.l r4,@r2

        ; save previous controller status
        mov.l @(4,r2),r4
        mov.l r4,@(8,r2)

        ; poll controller
        add #8,r2
        imov.l $20100075,r4
        mov #96,r0
        mov.b r0,@r4
@wastetime1:
        add #-2,r0
        cmp/eq #32,r0
        bf @wastetime1
        mov.b @r4,r1
        mov.b r1,@-r2           ; store 1st byte of controller data
        mov.b r0,@r4
@wastetime2:
        add #1,r0
        cmp/eq #64,r0
        bf @wastetime2
        mov.b @r4,r1
        mov.b r1,@-r2           ; store 2nd byte of controller data
        mov.b r0,@r4
@wastetime3:
        add #-2,r0
        cmp/eq #0,r0
        bf @wastetime3
        mov.b @r4,r1
        mov.b r1,@-r2           ; store 3rd byte of controller data
        mov.b r0,@r4
@wastetime4:
        add #1,r0
        cmp/eq #32,r0
        bf @wastetime4
        mov.b @r4,r1
        mov.b r1,@-r2           ; store 4th byte of controller data

        ; detect button presses (falling edges)
        mov.l @(4,r2),r4
        mov.l @r2,r1
        not r1,r1
        and r1,r4 
        mov.l r4,@(8,r2)

        ; update stuff in 68k memory
        add #-4,r2
        imov.l $5A00404,r1
        mov.l @r2,r4
        mov.l r4,@r1
        mov.l @(12,r2),r4
        mov.l r4,@(4,r1)

        ; acknowledge interrupt
        imov.l $25FE00A0,r2
        imov.l $FFFFBFFE,r4
        mov.l r4,@r2

        mov.l @r15+,r4
        mov.l @r15+,r2
        mov.l @r15+,r1
        mov.l @r15+,r0
        rte
        nop


        idata

main:

; init vdp2 registers
        imov.l $5F80000,r2    ; VDP2 registers address
        mov.l @(blah3,PC),r3    ; table with register values address
        imov.l $5F800FE,r5    ; a reserved register address to skip
        imov.l $5F80120,r6    ; last register + 2

        mov.w @r3+,r1
        mov.w r1,@r2            ; set $5F80000 and then skip to $5F8000E
        add #$0E,r2

loop1:
        mov.w @r3+,r1
        cmp/eq r2,r5
        bt nowrite
        mov.w r1,@r2
nowrite:
        add #2,r2
        cmp/eq r2,r6
        bf loop1


; init controller port
        imov.l $20100075,r1
        mov #0,r0
        mov.b r0,@(10,r1)       ; clear EXLE
        mov #3,r0
        mov.b r0,@(8,r1)        ; set IOSEL
        mov #$60,r0
        mov.b r0,@(4,r1)        ; set bits 5,6 to output in DDR1/DDR2
        mov.b r0,@(6,r1)        ; set bits 5,6 to output in DDR1/DDR2

; init SCSP
        imov.l $20F,r0
        imov.l $5B00400,r1
        mov.w r0,@r1

; copy 68k code to 68k RAM
        mov.l @(playeraddr,PC),r0
        imov.l $5A10000,r6    ; last word write address + 2
        imov.l $5A00000,r2
@loop3:
        mov.w @r0+,r1
        mov.w r1,@r2
        add #2,r2
        cmp/eq r2,r6
        bf @loop3
; copy samples to 68k RAM
        mov.l @(samplesaddr,PC),r0
        imov.l $5A50000,r6    ; last word write address + 2
@loop5:
        mov.w @r0+,r1
        mov.w r1,@r2
        add #2,r2
        cmp/eq r2,r6
        bf @loop5

        ; turn on 68000
        mov #6,r0
        imov.l $2010001F,r1
        mov.b r0,@r1

        ; wait for handshake
;        imov.l $FFFFABCD,r1
;        imov.l $25A00400,r2
;@loop4:
;        mov.w @r2,r0
;        cmp/eq r0,r1
;        bf @loop4


        bra copybmps
        nop

        idata
        
blah3:
        dc.l regvalues2

playeraddr:
        dc.l player
samplesaddr:
        dc.l samples


; register settings for a 352x224 screen with NBG0 as a 32K color bitmap
; and NBG1

regvalues2:
        dc.w $8081
        dc.w 0
        dc.w $EF44,$EF44,$FFFF,$FFFF,$EF55,$EF55,$FFFF,$FFFF
;        dc.w $0101,0,0,0,$0036,0,0,0
        dc.w $0103,0,0,0,$3232,0,0,0
        dc.w 0,0,0,0,0,0,$20,0
        dc.w 0,0,0,0,0,0,0,0
        dc.w 0,0,0,0,0,0,0,0
        dc.w 0,0,0,0,0,0,0,0
        dc.w 0,0,0,0,1,0,1,0
        dc.w 0,0,0,0,1,0,1,0
        dc.w 0,0,0,0,0,0,0,0
        dc.w 0,0,0,0,0,0,$8003,$C000
        dc.w 0,0,0,0,0,0,0,0
        dc.w 0,0,0,0,0,0,0,0
        dc.w 0,0,0,0,0,0,0,0
;        dc.w $7020,0,0,0,0,0,$441,0
        dc.w $3020,0,0,0,0,0,$41,0
;        dc.w $707,$707,$707,$707,6,0,0,0
        dc.w $505,$505,$505,$505,$706,0,0,0
        dc.w $1010,$1010,$1010,$1010,$18,0,0,0
        dc.w 0,0,0,0,0,0,0,0


; 5E00000-5E7FFFF - VDP2 RAM
; 5F00000-5F00FFF - color RAM
; vdp 2 registers
; 5F80000 r/w - screen mode
;       bit 15 disp - 1=enabled
;       bit 8 bdclmd - 0=black, 1=back screen
;       bit 7-6 lsmd - 0=normal, 2=flickery low-res, 3=interlaced high-res
;       bits 5-4 vreso - 0=224 lines, 1=240 lines
;       bit 2 hreso2 - 0=normal, 1=31khz horizontal frequency
;       bit 1 hreso1 - 0=normal, 1=double hres
;       bit 0 hreso0 - 0=320/640 pixels, 1=352/704 pixels
; 5F80002 r/w - external signal enable
;       bit 9 exlten - 0=latch h/v counters when reading this register
;                       1=latch on external signal (eg. light gun)
;       bit 8 exsyen - 0=normal, 1=external sync
;       bit 1 dasel - (something to do with external video priority?)
;       bit 0 exbgen - 1=substitute external video for nbg1
; 5F80004 r - screen status
;       bit 9 exltfg - 1=hv counters are latched, resets on read
;       bit 8 exsyfg - set when vdp is synchronized with external signal
;       bit 3 vblank
;       bit 2 hblank
;       bit 1 odd - (always shows odd in non-interlaced mode)
;       bit 0 pal
; 5F80006 r/w - vram size
;       bit 15 vramsz - 0=4mbit 1=8mbit
;       bits 3-0 vdp version
; 5F80008 r - h-counter bits 9-0 - (counts 14mhz pixels)
; 5F8000A r - v-counter bits 9-0 - counts 0-447
;       bit 0 clear for non-interlace 15khz, set for interlaced even fields
; 5F8000C reserved
; 5F8000E r/w - RAM control
;       bit 15 crkte - 0=coefficient table stored in VRAM, 1=stored in CRAM
;       bits 13-12 crmd - 0=RGB15x1024, 1=RGB15x2048, 2=RGB24x1024
;       bits 9 vrbmd - \
;       bits 8 vramd -  \ 1=partition into two banks
;       bits 7-6 rdbsa0 - \    0=not used for RBG0
;       bits 5-4 rdbsa1 -  \   1=used for end table
;       bits 3-2 rdbsb0 -   \  2=used for name table
;       bits 1-0 rdbsb1 -    \ 3=used for character set (or bitmap)
; 5F80010 w - vram cycle pattern A0 (nibbles 0-3)
; 5F80012 w - vram cycle pattern A0 (nibbles 4-7)
; 5F80014 w - vram cycle pattern A1 values 0-3 NBGx name table read
; 5F80016 w - vram cycle pattern A1        4-7 NBGx charset/bitmap read
; 5F80018 w - vram cycle pattern B0        8-11 undefined
; 5F8001A w - vram cycle pattern B0        12-13 NBG0/1 vscroll table read
; 5F8001C w - vram cycle pattern B1        14 CPU access
; 5F8001E w - vram cycle pattern B1        15 no access
; 5F80020 w - screen display enable
;       bits 12-8 r0/n3/n2/n1/n0 transparency disable
;       bits 5-0 r1/r0/n3/n2/n1/n0 display enable
; 5F80022 w - mosaic control
;       bits 15-12 mzszv - vertical dot size (1-16 pixels)
;       bits 11-18 mzszh - horizontal dot size (1-16 pixels)
;       bits 4-0 - r0/n3/n2/n1/n0 mosaic enable
; 5F80024 w - special function code select
;       bits 4-0 - r0/n3/n2/n1/n0 - 0=code A, 1=code B
; 5F80026 w - special function code
;       bits 15-8 sfcdb         bits 7-0 sfcda
;               enables a special function for certain color values
;       setting bit 0 - activates when color value bits 1-3 are 0
;       setting bit 1 - activates when color value bits 1-3 are 1
;       ...
;       setting bit 7 - activates when color value bits 1-3 are 7
;
; 5F80028 w - character control (n0,n1)
;       bits 13-12 - nbg1 color depth 
;       bits 11-10 - nbg1 bitmap size
;       bit 9 - nbg1 bitmap enable      bit 8 - nbg1 character size
;       bits 6-4 - nbg0 color depth 
;       bits 3-2 - nbg0 bitmap size
;       bit 1 - nbg0 bitmap enable      bit 0 - nbg0 character size
; 5F8002A w - character control (n2,n3,r0)
;       bits 14-12 - rbg0 color depth
;       bits 10 - rbg0 bitmap size
;       bit 9 - rbg0 bitmap enable      bit 8 - rbg0 character size
;       bit 5 - nbg3 color depth        bit 4 - nbg3 character size
;       bit 1 - nbg2 color depth        bit 0 - nbg2 character size
;         color depths  0=16          bitmap sizes  0=512x256
;                       1=256                       1=512x512
;                       2=2048                      2=1024x256
;                       3=32768                     3=1024x512
;                       4=16777216    char sizes 0=8x8, 1=16x16
; 5F8002C w - bitmap palette number (nbg0,nbg1)
;       bit 13 - priority               bit 12 - color calculation enable
;       bits 10-8 - nbg1 bitmap palette
;       bit 5 - priority               bit 4 - color calculation enable
;       bits 2-1 - nbg0 bitmap palette
; 5F8002E w - bitmap palette number (rbg0)
;       bit 5 - priority               bit 4 - color calculation enable
;       bits 2-1 - rbg0 bitmap palette
; 5F80030 w - pattern name control nbg0
; 5F80032 w - pattern name control nbg1
; 5F80034 w - pattern name control nbg2
; 5F80036 w - pattern name control nbg3
; 5F80038 w - pattern name control rbg0
;       bit 15 - name data size 0=1 word, 1=2 words
;       bit 14 - 0=10bit char number with h/v flip, 1=12bit char number
;       bit 9 - priority something-or-other
;       bit 8 - color calculation something-or-other
;       bits 7-5 - palette number (top 3 bits?)
;       bits 4-0 - charset number?
; 5F8003A w - plane size
;       bits 15-14 rotation parameter A screen-over process\  0=repetition
;       bits 13-12 rotation parameter A plane size          | 1=character
;       bits 11-10 rotation parameter B screen-over process/  2=transparent
;       bits 9-8 rotation parameter B plane size
;       bits 7-6 NBG3 plane size \
;       bits 5-4 NBG2 plane size  \   0=1 page
;       bits 3-2 NBG1 plane size   \  1=2 pages x 1 page
;       bits 1-0 NBG0 plane size    \ 2=2 pages x 2 pages
; 5F8003C w - map offset (also bitmap address)
;       bits 14-12 NBG3  \   0=0
;       bits 10-8 NBG2    \  1=$20000
;       bits 6-4 NBG1     /  2=$40000
;       bits 2-0 NBG0    /   3=$60000
; 5F8003E w - map offset
;       bits 6-4 rotation parameter B
;       bits 2-0 rotation parameter A
; 5F80040 w - map NBG0 - bits 13-8 plane B, bits 5-0 plane A
; 5F80042 w - map NBG0 - bits 13-8 plane D, bits 5-0 plane C
; 5F80044 w - map NBG1 - bits 13-8 plane B, bits 5-0 plane A
; 5F80046 w - map NBG1 - bits 13-8 plane D, bits 5-0 plane C
; 5F80048 w - map NBG2 - bits 13-8 plane B, bits 5-0 plane A
; 5F8004A w - map NBG2 - bits 13-8 plane D, bits 5-0 plane C
; 5F8004C w - map NBG3 - bits 13-8 plane B, bits 5-0 plane A
; 5F8004E w - map NBG3 - bits 13-8 plane D, bits 5-0 plane C
; 5F80050 w - map rotation A - bits 13-8 plane B, bits 5-0 plane A
; 5F80052 w - map rotation A - bits 13-8 plane D, bits 5-0 plane C
; 5F80054 w - map rotation A - bits 13-8 plane F, bits 5-0 plane E
; 5F80056 w - map rotation A - bits 13-8 plane H, bits 5-0 plane G
; 5F80058 w - map rotation A - bits 13-8 plane J, bits 5-0 plane I
; 5F8005A w - map rotation A - bits 13-8 plane L, bits 5-0 plane K
; 5F8005C w - map rotation A - bits 13-8 plane N, bits 5-0 plane M
; 5F8005E w - map rotation A - bits 13-8 plane P, bits 5-0 plane O
; 5F80060 w - map rotation B - bits 13-8 plane B, bits 5-0 plane A
; 5F80062 w - map rotation B - bits 13-8 plane D, bits 5-0 plane C
; 5F80064 w - map rotation B - bits 13-8 plane F, bits 5-0 plane E
; 5F80066 w - map rotation B - bits 13-8 plane H, bits 5-0 plane G
; 5F80068 w - map rotation B - bits 13-8 plane J, bits 5-0 plane I
; 5F8006A w - map rotation B - bits 13-8 plane L, bits 5-0 plane K
; 5F8006C w - map rotation B - bits 13-8 plane N, bits 5-0 plane M
; 5F8006E w - map rotation B - bits 13-8 plane P, bits 5-0 plane O
;
; 5F80070 w - NBG0 horizontal scroll (integer) bits 10-0
; 5F80072 w - NBG0 horizontal scroll (fraction) bits 15-8
; 5F80074 w - NBG0 vertical scroll (integer) bits 10-0
; 5F80076 w - NBG0 vertical scroll (fraction) bits 15-8
; 5F80078 w - NBG0 horizontal increment (integer) bits 2-0
; 5F8007A w - NBG0 horizontal increment (fraction) bits 15-8
; 5F8007C w - NBG0 vertical increment (integer) bits 2-0
; 5F8007E w - NBG0 vertical increment (fraction) bits 15-8
; 5F80080 w - NBG1 horizontal scroll (integer) bits 10-0
; 5F80082 w - NBG1 horizontal scroll (fraction) bits 15-8
; 5F80084 w - NBG1 vertical scroll (integer) bits 10-0
; 5F80086 w - NBG1 vertical scroll (fraction) bits 15-8
; 5F80088 w - NBG1 horizontal increment (integer) bits 2-0
; 5F8008A w - NBG1 horizontal increment (fraction) bits 15-8
; 5F8008C w - NBG1 vertical increment (integer) bits 2-0
; 5F8008E w - NBG1 vertical increment (fraction) bits 15-8
; 5F80090 w - NBG2 horizontal scroll (integer) bits 10-0
; 5F80092 w - NBG2 vertical scroll (integer) bits 10-0
; 5F80094 w - NBG3 horizontal scroll (integer) bits 10-0
; 5F80096 w - NBG3 vertical scroll (integer) bits 10-0
;
; 5F80098 w - horizontal reduction enable
;       bits 9-8 nbg1 \  0=normal       1=half size
;       bits 1-0 nbg0  \ 2 or 3=quarter size
; 5F8009A w - line and vertical cell scroll control (??)
; 5F8009C-E w - vertical cell scroll table address (bits 18-1) / VRAM
; 5F800A0-2 w - NBG0 line scroll table address (bits 18-1)    / address
; 5F800A4-6 w - NBG1 line scroll table address (bits 18-1)   / divided by 2
; 5F800A8-A w - line color table address (bits 18-1) bit 31=lct enable
;       (line color table is used only for color calculation
;        and if enabled, has highest priority)
; 5F800AC-E w - back screen table address (bits 18-1) bit 31=bst enable
;       (back screen has lowest priority)
; 5F800B0 w - rotation parameter mode rbg0
;                0=A    2=according to coefficient data
;                1=B    3=according to rotation parameter window
; 5F800B2 w - rotation parameter read control (??)
; 5F800B4 w - coefficient table control ...
;
; ...
;
; 5F800E0 w - sprite control
;               bits 13-12 color calculation condition
;                       0=priority<=color calc number
;                       1=priority=color calc number
;                       2=priority>=color calc number
;                       3=disabled
;                       (only applies to pixels with MSB=0, ie. not RGB)
;               bits 10-8 color calculation condition number
;               bit 5 sprite color mode 0=palettized only 1=palettized and RGB
;               bit 4 sprite window enable
;               bits 3-0 sprite type (?)
;
; ...
;
; 5F800EC w - color calculation control register
;               bit 15 1=enable color gradiant function
;               bits 14-12 gradiant screen select
;                       0 = sprite layer        1 = RBG0
;                       2 = NBG0/RBG1           3 = invalid
;                       4 = NBG1/EXBG           5 = NBG2
;                       6 = NBG2                7 = invalid
;               bit 10 1=enable extended color calculation mode
;       (extended mode allows color addition of up to 4 layers)
;       (can't be used with hi-res/31khz or with gradiant function)
;               bit 9 CC ratio mode 0=depends on top layer 1=depends on bottom
;               bit 8 CC addition mode 0=depends on CCR registers 1=as is
;               bit 6 sprite layer CC enable
;               bit 5 LNCL layer CC enable
;               bit 4 RBG0 layer CC enable
;               bit 3 NBG3 layer CC enable
;               bit 2 NBG2 layer CC enable
;               bit 1 NBG1/EXBG layer CC enable
;               bit 0 NBG0/RBG1 layer CC enable
; 5F800EE w - special color calculation mode
;               bits 9-8 RBG0      \   0=enabled for entire layer
;               bits 7-6 NBG3       \  1=enabled per character
;               bits 5-4 NBG2        | 2=enabled per dot
;               bits 3-2 NBG1/EXBG  /  3=enabled by color MSB
;               bits 1-0 NBG0/RBG1 /
;       (applies to top-most layer of color calculation only)               
;
; 5F800F0 w - priority - bits 10-8 sprite type 1, bits 2-0 sprite type 0
; 5F800F2 w - priority - bits 10-8 sprite type 3, bits 2-0 sprite type 2
; 5F800F4 w - priority - bits 10-8 sprite type 5, bits 2-0 sprite type 4
; 5F800F6 w - priority - bits 10-8 sprite type 7, bits 2-0 sprite type 6
; 5F800F8 w - priority number - bits 10-8 NBG1, bits 2-0 NBG0
; 5F800FA w - priority number - bits 10-8 NBG3, bits 2-0 NBG2
; 5F800FC w - priority number - bits 2-0 RBG0
; 5F800FE reserved
;       (changing priorities mid-screen has no immediate effect?)
;
; 5F80100 w - color calculation ratio register
;               bits 12-8 sprite type 1, bits 4-0 sprite type 0
; 5F80102 w - color calculation ratio register
;               bits 12-8 sprite type 3, bits 4-0 sprite type 2
; 5F80104 w - color calculation ratio register
;               bits 12-8 sprite type 5, bits 4-0 sprite type 4
; 5F80106 w - color calculation ratio register
;               bits 12-8 sprite type 7, bits 4-0 sprite type 6
;
; 5F80108 w - color calculation ratio register 
;               bits 12-8 NBG1/EXBG     (value is fraction x/32)
;               bits 4-0 NBG0/RBG1
; 5F8010A w - color calculation ratio register 
;               bits 12-8 NBG3          (value is fraction x/32)
;               bits 4-0 NBG2     
; 5F8010C w - color calculation ratio register 
;               bits 4-0 RBG0
; 5F8010E w - color calculation ratio register 
;               bits 12-8 back screen   (value is fraction x/32)
;               bits 4-0 LNCL 
; 5F80110 w - color offset enable register
;               bit 6 sprite            bit 5 backscreen
;               bit 4 RBG0              bit 3 NBG3
;               bit 2 NBG2              bit 1 NBG1/EXBG
;               bit 0 NBG0/RBG1
; ...

clearbmp:
        mov.w r0,@r1
        add #2,r1
        cmp/eq r1,r2
        bf clearbmp
        rts
        nop

copybmp:
        mov.w @r0+,r2   ; get width
        mov.w @r0+,r3   ; get height

        add #4,r0

        mov.l @(blahbh,PC),r9   ; screen bitmap width in bytes

        mov -1,r8

        mov #0,r7
loopy2:
        mov r1,r4
        add r9,r1
        mov #0,r6
loopy:
        cmp/pz r8
        bt a112

        mov.w @r0+,r8
        bra a113
        mov.w r8,@r4

a112:
        ; awesomest data compression scheme ever...
        ; if the high bit was clear then write the pixel twice
        mov.w r8,@r4
        mov -1,r8
a113:
        add #2,r4
        add #1,r6
        cmp/eq r6,r2
        bf loopy
        add #1,r7
        cmp/eq r7,r3
        bf loopy2
        rts
        nop

copybmps:
        mov.l @(bmpaddr,PC),r0
        imov.l $5E00000,r1    ; VDP2 RAM address
        bsr copybmp
        nop

;        mov.l @(bmpaddr2,PC),r0
        imov.l $5E40000,r1    ; VDP2 RAM address
;        bsr copybmp
;        nop
        imov.l $5E78000,r2
        mov #0,r0
        bsr clearbmp
        nop

        ; copy backscreen table
        mov.l @(bsaddr,PC),r0
        imov.l $5E78000,r1    ; VDP2 RAM address
        imov.l #224,r2
@bloop:
        mov.w @r0+,r3
        mov.w r3,@r1
        add #2,r1
        dt r2
        bf @bloop

        bra vdp1crap
        nop

        idata

        align.l
blahd:
        dc.l $05E00000
blahbh:
        dc.l 1024
bmpaddr:
        dc.l thebmp
;bmpaddr2:
;        dc.l thebmp2
bsaddr:
        dc.l backscreen

vdp1crap:
; init vdp1 command tables
        mova @(ptable,PC),r0        ; table with register values
        imov.l $5C00000,r2    ; VDP1 RAM base address
        imov.l $5C000E0,r6    ; last word write address + 2

@loop2:
        mov.w @r0+,r1
        mov.w r1,@r2
        add #2,r2
        cmp/eq r2,r6
        bf @loop2


; init vdp1 registers
        mova @(regvalues1,PC),r0        ; table with register values
        imov.l $5D00000,r2    ; VDP1 registers address
        imov.l $5D0000C,r6    ; last register + 2

@loop1:
        mov.w @r0+,r1
        mov.w r1,@r2
        add #2,r2
        cmp/eq r2,r6
        bf @loop1

; copy textures
        mov.l @(texaddr,PC),r0
        imov.l $5C40000,r2    ; VDP1 RAM base address
        imov.l $5C80000,r6    ; last word write address + 2

@loop3:
        mov.w @r0+,r1
        mov.w r1,@r2
        add #2,r2
        cmp/eq r2,r6
        bf @loop3

        bra rendercrap
        nop

        idata

texaddr:
        dc.l atexture

regvalues1:
        dc.w 0,0,2,0
        dc.w 0,$58E0


; 5C00000-5C7FFFF - VDP1 work RAM
; 5C80000-5CBFFFF - off-screen frame buffer access
; vdp 1 registers
; 5D00000 w - tv mode
;               bit 3 enables v-blank erase
;               bits 0-2 display mode
;                       0=normal 512x256        1=high res 1024x256
;                       2=rotated 512x256       3=rotated 512x512
;                       4=31KHz
; 5D00002 w - frame buffer change register
;               bit 4 even/odd coordinate select bit 0=even 1=odd
;                       determines which pixels are sampled when drawing
;                       scaled/distorted sprites using high-speed shrink mode
;               bit 3 double interlace enable
;               bit 2 - set to 0 for normal mode
;                       when in double interlace mode 0=even lines 1=odd
;                       (even and odd lines are rendered to different buffers)
;               bits 0-1 change mode
;                       0=automatic - switches buffers and erases
;                       2=manual - no change
;                       3=reverts to manual after one change+erase
; 5D00004 w - plot trigger register 0=idle 1=start now 2=start at frame change
; 5D00006 w - data word to be written during frame buffer erase
; 5D00008 w - erasure upper left coordinate
;               bits 14-9 - X coordinate (units of 16 bytes)
;               bits 8-0 - Y coordinate
; 5D0000A w - erasure lower right coordinate
;               bits 15-9 - X coordinate (units of 16 bytes)
;               bits 8-0 - Y coordinate
; 5D0000C w - write zero to stop drawing
; 5D00010 r - status register
;               bit 1 - set if drawing is done
;               bit 0 - set if drawing was done at the end of previous frame
; 5D00012 r - last operation command address (at previous frame end)
; 5D00014 r - current operation command address
; 5D00016 r - mode status register (mirrors some bits from first 3 registers)


ptable:
        dc.w 9,0,0,0,0,0,0,0
;        dc.w 0,0,351,223,0,0,0,0
        dc.w 0,0,351,447,0,0,0,0

;        dc.w 10,0,0,0,0,0,176,112
        dc.w 10,0,0,0,0,0,176,224
        dc.w 0,0,0,0,0,0,0,0

        dc.w $8002,0,$08E8,$FFFF,$8000,$0410,40,40
        dc.w 59,40,59,59,40,59,0,0



; command table format
; 00 - control word
;       bit 15 - end drawing (stops without processing this table)
;       bits 14-12 jump mode
;               0=advances to the next table after this one
;               1=jumps to CMDLINK table after this one
;               2=CMDLINK table receives subroutine call after this table
;               3=returns from subroutine after this table
;               4=skips to the next table
;               5=jumps to CMDLINK table, skipping this one
;               6=CMDLINK table receives subroutine call, skipping this one
;               7=returns from subroutine without processing this table
;       bits 11-8 zoom (point which stays fixed when a sprite is scaled)
;               0=none (uses vertex A and vertex C as corners)
;               5=upper left    6=upper center  7=upper right
;               9=centerleft   10=centercenter 11=centerright
;              13=lower left   14=lower center 15=lower right
;                       (uses vertex A as fixed point, vertex B as size)
;       bits 5-4 character read direction
;               0=normal        1=H flip
;               2=V flip        3=H+V flip
;       bits 3-0 command select
;               0=normal sprite         1=scaled sprite
;               2=distorted sprite     
;               4=polygon               5=polyline
;               6=line
;               8=set user clipping     9=set system clipping
;              10=set local coordinate
; 02 - CMDLINK
;       bits 15-2 address of command table - used for jumps/calls
; 04 - draw mode word
;       bit 15 - sets MSB for all pixels that are written
;       bit 12 - enables high-speed shrink
;       bit 11 - pre-clipping disable
;       bit 10 - user clipping enable (system clipping is always on)
;       bit 9 - specifies whether clipping is done inside or outside
;               0=drawing is done inside
;       bit 8 - mesh enable
;       bit 7 - end code disable (character pattern)
;       bit 6 - transparent pixel disable
;       bits 5-3 color mode
;               0=16 colors palettized  1=16 colors VDP1 lookup table
;               2=64 colors palettized  3=128 colors palettized
;               4=256 colors palettized 5=32K colors
;       bit 2 - gouraud shading enable
;       bit 1 - 1/2 luminance enable           \
;       bit 0 - background 1/2 luminance enable \ (combine for transparency)
; 06 - aux color
;       specifies either a color bank, lookup table address, or RGB color
;                         (top 8-12 bits)    (top 14 bits)
; 08 - character address (top 14 bits) (write bottom 2 bits as 0)
; 0A - character size (bits 13-8 are X/8, bits 7-0 are Y)
; 0C - vertex A - X coordinate \
; 0E - vertex A - Y coordinate  \
; 10 - vertex B - X coordinate   \
; 12 - vertex B - Y coordinate    \ value ranges from -1024 to 1023
; 14 - vertex C - X coordinate    / and is sign-extended to 16 bits
; 16 - vertex C - Y coordinate   /
; 18 - vertex D - X coordinate  /
; 1A - vertex D - Y coordinate /
; 1C - gouraud shading table address (all 16 bits)
; system clipping uses vertex C (as lower right corner)
; user clipping uses vertex A and vertex C
; local coordinates uses vertex A


; Saturn SH-2 memory map (add $20000000 for non-cacheing access)
;
; 0-1FFFF boot ROM
; 100000 SMPC (registers in this range are written by BIOS routines)
; 180000-18FFFF backup RAM?
; 200000-2FFFFF work RAM low?
; 2000000-3FFFFFF cartridge
;
; 5A00000-5A7FFFF sound RAM
; 5B00000 SCSP registers
; 5C00000-5C7FFFF VDP1 RAM
; 5C80000-5CBFFFF off-screen frame buffer
; 5D00000 VDP1 registers
; 5E00000-5E7FFFF VDP2 RAM
; 5F00000-5F00FFF color RAM
; 5F80000 VDP2 registers
; 5FC0000 SMPC?
; 5FE0000 SCU
;
; 6000000-60FFFFF SDRAM


rendercrap:

        imov.l $60DF000,r14    ; "zero page"

        mov.l @(textaddr,PC),r4
        mov.l r4,@($34,r14)     ; init sometext pointer
        mov.l @(scriptaddr,PC),r4
        mov.l r4,@($38,r14)     ; init apscript pointer

        imov.l $5C00040,r3    ; command table address for buildpolys
        mov.l r3,@($10,r14)

        mov #0,r4
        mov.l r4,@($14,r14)     ; "player" position
        mov.l r4,@($0C,r14)     ; orientation (index into sine table)
        mov.l r4,@($30,r14)     ; movement
        mov.l r4,@($3C,r14)     ; script timer
        imov.l $50005,r0
        mov.l r0,@($2C,r14)     ; map location

buildtunnel:
        ; build a model of the tunnel
        imov.l $60E0000,r8     ; vertex buffer
        imov.l $60F8000,r9     ; polygon list
        imov.l $60EC000,r12    ; polygon attributes table

        mov #0,r10      ; number of verteces
        mov #0,r11      ; number of polygons

        ; get position on dungeon map
        mov.l @($2C,r14),r0
        swap.w r0,r1
        mov #15,r2
        and r2,r1       ; Y (or Z, depending on how you look at it...)
        and r0,r2       ; X

        ; render sectors starting 3 blocks out, working inward
        mov.l @(sladdr,PC),r5
        mov #25,r6
a139:
        mov.b @r5+,r3           ; dY
        mov.b @r5+,r4           ; dX
;        mov r3,r7
        neg r3,r7
        shll8 r7
        mov.w @($14,r14),r0
;        sub r0,r7
        add r0,r7
        mov.l r7,@($18,r14)        ; rendering sector z position
        mov r4,r7
        shll8 r7
        mov.w @($16,r14),r0
        sub r0,r7
        mov.l r7,@(8,r14)        ; rendering sector x position
        add r1,r3
        cmp/pz r3
        bf a140
        add r2,r4
        cmp/pz r4
        bf a140
        ; calculate address and get dungeon map word
        mov.l @(mapaddr,PC),r7
        shll2 r3
        shll2 r3
        add r3,r4
        shll r4
        add r4,r7
        mov.w @r7,r0

        mov.l r1,@-r15
        mov.l r2,@-r15
        mov.l r5,@-r15
        mov.l r6,@-r15
        bsr renderwalls
        nop
        mov.l @r15+,r6
        mov.l @r15+,r5
        mov.l @r15+,r2
        mov.l @r15+,r1

a140:
        dt r6
        bf a139

        mov.l r10,@r14          ; number of verteces
        mov.l r11,@(4,r14)      ; number of polygons

        bra transforms
        nop


        align.l
scriptaddr:
        dc.l apscript
textaddr:
        dc.l sometext
sladdr:
        dc.l searchlist
mapaddr:
        dc.l dungeon

        idata

renderwalls:
        sts.l PR,@-r15

        tst #4,r0       ; left wall
        bt a135
        imov.l #-128,r3         ; x coord
        bsr sidewall
        nop
a135:

        tst #8,r0       ; right wall
        bt a136
        imov.l #128,r3         ; x coord
        bsr sidewall
        nop
a136:

        tst #1,r0       ; ahead wall
        bt a137
        imov.l #128,r3         ; depth
        bsr endwall
        nop
a137:

        tst #2,r0       ; behind wall
        bt a138
        imov.l #-128,r3         ; depth
        bsr endwall
        nop
a138:

        lds.l @r15+,PR
        rts
        nop


sidewall:
        mov.l r0,@-r15
        mov.l @(8,r14),r4      ; sector x
        add r4,r3
        mov r10,r5      ; starting vertex number
        mov #-90,r4         ; y coord 1
        mov #2,r2       ; number of rows
a121:
        mov.l @($18,r14),r13      ; depth
        add #-64,r13
        add #-64,r13
        mov #5,r1       ; number of columns
a120:
        mov.l r3,@r8
        mov.l r4,@(4,r8)
        mov.l r13,@(8,r8)
        add #12,r8
        add #64,r13
        add #1,r10
        dt r1
        bf a120
        mov #90,r4         ; y coord 2
        dt r2
        bf a121
a123:
        ; add polygons
        mov #4,r1       ; number of columns
        imov.l $81408000,r6
        imov.l $0250,r7
a122:
        ; assign verteces to polygon
        mov.w r5,@r9
        mov r5,r0
        add #1,r0
        mov.w r0,@(2,r9)
        add #5,r0
        mov.w r0,@(4,r9)
        add #-1,r0
        mov.w r0,@(6,r9)
        add #8,r9
        add #1,r5
        add #1,r11
        ; assign texture
        mov.w r6,@r12
        swap.w r6,r6
        add #2,r12
        mov.w r7,@r12
        add #2,r12
        dt r1
        bf a122
        mov.l @r15+,r0
        rts
        nop

endwall:
        mov.l r0,@-r15
        mov.l @($18,r14),r13      ; depth
        add r3,r13
        mov r10,r5      ; starting vertex number
        mov #-90,r4         ; y coord 1
        mov #2,r2       ; number of rows
a125:
        mov.l @(8,r14),r3          ; sector x
        add #-64,r3
        add #-64,r3
        mov #5,r1       ; number of columns
a124:
        mov.l r3,@r8
        mov.l r4,@(4,r8)
        mov.l r13,@(8,r8)
        add #12,r8
        add #64,r3
        add #1,r10
        dt r1
        bf a124
        mov #90,r4         ; y coord 2
        dt r2
        bf a125
        bra a123
        nop


        idata


transforms:

        ; perform lookup of current viewing angle sin, cos, -sin

        imov.l $3FFC,r8
        imov.l $1000,r9
        mov.l @(stadr,PC),r6
        mov r14,r13
        add #$40,r13

        mov.w @($0C,r14),r0
        shlr2 r0
        and r8,r0
        mov.l @(r0,r6),r7
        mov.l r7,@r13           ; sin
        add r9,r0
        and r8,r0
        mov.l @(r0,r6),r7
        mov.l r7,@(4,r13)     ; cos
        add r9,r0
        and r8,r0
        mov.l @(r0,r6),r7
        mov.l r7,@(8,r13)     ; -sin


        ; process the entire list of verteces
        mov.l @r14,r1
        imov.l $60E0000,r2
        imov.l $60F0000,r3

a102:
        mov r14,r13
        add #$40,r13

        ; rotate points on X-Z plane
        mov.l @r2+,r5           ; X
        mov.l @r13+,r7          ; sin
        dmuls.l r5,r7           ; X*sin
        mov.l @r2+,r10          ; (Y remains unchanged)
        mov.l @r13,r11          ; (grab cos for later)
        mac.l @r2+,@r13+        ; +Z*cos
        sts MACL,r6
        shlr16 r6
        exts.w r6,r6            ; new Z        

        dmuls.l r5,r11          ; X*cos
        add #-4,r2
        mac.l @r2+,@r13+        ; +Z*-sin
        sts MACL,r5
        shlr16 r5
        exts.w r5,r5            ; new X        


;        mov.l @r2+,r5
;        mov.l @r2+,r10
;        mov.l @r2+,r6

        ; clip Z
        mov #89,r0
        cmp/gt r0,r6
        bt a130
        mov #90,r6
a130:

;        add #10,r6      ; debugging

        imov.l #176,r7     ; zoom H
        dmuls.l r7,r5
        sts MACL,r5

;        imov.l #144,r7     ; zoom V
        imov.l #288,r7     ; zoom V
        dmuls.l r7,r10
        sts MACL,r10

        ; divide x by distance
        mov r6,r7
        shll16 r7
        mov #0,r8
        mov r5,r9
        rotcl r9
        subc r8,r5
        div0s r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        div1 r7,r5
        exts.w r5,r5
        rotcl r5
        addc r8,r5

        ; divide y by distance
        mov r6,r7
        shll16 r7
        mov #0,r8
        mov r10,r9
        rotcl r9
        subc r8,r10
        div0s r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        div1 r7,r10
        exts.w r10,r10
        rotcl r10
        addc r8,r10

a104:
        ; clip X negative
        imov.l #-223,r0
        cmp/gt r0,r5
        bf a154
        ; clip X positive
        imov.l #223,r0
        cmp/gt r5,r0
        bt a155
a154:
        mov r0,r5
a155:

        ; store transformed x,y,z
        mov.w r5,@r3
        add #2,r3
        mov.w r10,@r3
        add #2,r3
        mov.w r6,@r3
        add #4,r3

        dt r1
        bf a102

        bra buildpolys
        nop

        idata


buildpolys:

        ; build VDP1 command tables
        imov.l $60F0000,r4

        mov.l @(4,r14),r1
        mov.l @($10,r14),r3    ; command table write address
        imov.l $60F8000,r2     ; polygon data table
        imov.l $60EC000,r12    ; polygon attributes table

a103:
        ; get texture assignment
        mov.l @r12+,r0
        mov.l r0,@(8,r3)   ; texture address and size

        ; verteces
        mov.w @r2+,r5
        shll r5
        shll2 r5
        add r4,r5
        mov.w @r5+,r0
        mov r0,r7
        mov.w r0,@(12,r3)
        mov.w @r5+,r0
        mov.w r0,@(14,r3)
        mov.w @r5+,r6

        mov.w @r2+,r5
        shll r5
        shll2 r5
        add r4,r5
        mov.w @r5+,r0
        add r0,r7
        mov.w r0,@(16,r3)
        mov.w @r5+,r0
        mov.w r0,@(18,r3)
        mov.w @r5+,r0
        add r0,r6

        mov.w @r2+,r5
        shll r5
        shll2 r5
        add r4,r5
        mov.w @r5+,r0
        add r0,r7
        mov.w r0,@(20,r3)
        mov.w @r5+,r0
        mov.w r0,@(22,r3)
        mov.w @r5+,r0
        add r0,r6

        mov.w @r2+,r5
        shll r5
        shll2 r5
        add r4,r5
        mov.w @r5+,r0
        add r0,r7
        ; check sum of all four Xs to identify offscreen polygons (and skip)
        imov.l #-892,r8         ;
        cmp/eq r7,r8            ;
        bt a106                 ;
        neg r8,r8               ;
        cmp/eq r7,r8            ;
        bt a106                 ;
        mov.w r0,@(24,r3)
        mov.w @r5+,r0
        mov.w r0,@(26,r3)
        mov.w @r5+,r0
        add r6,r0

        ; check sum of all four Zs to identify offscreen polygons (and skip)
        shlr2 r0
        cmp/eq #90,r0
        bt a106
        
        mov #2,r0       ; 2=textures, 4=untextured, 5=lines
        mov.w r0,@r3            ; control word

        mov #0,r0
        mov.w r0,@(2,r3)        ; CMDLINK
        mov.w r0,@(28,r3)       ; shading table address

        imov.l $08E8F595,r0
        mov.l r0,@(4,r3)        ; draw mode/aux color

        add #$20,r3
a106:

        dt r1
        bf a103

        imov.l $8004,r0
        mov.w r0,@r3            ; end command

        ; switch command table addresses 
        imov.l $B802080,r1
        mov.l @($10,r14),r3
        sub r3,r1
        mov.l r1,@($10,r14)
        swap.b r1,r0
        tst #$20,r0
        bf a129
        ; insert a jump
        mov #$50,r0
        mov.b r0,@r1
        imov.l $408,r0
        mov.w r0,@(2,r1)
a129:

        ; draw some shit representing controller port bits (debugging)
;        imov.l $5E05020,r1
;        mov.l @($20,r14),r0
;        imov.l $80008000,r2
;        mov #-1,r3
;        imov.l $400,r4
;        mov #32,r5
;@debugloop:
;        tst #1,r0
;        bt @debug1
;        mov.l r2,@r1
;        bf @debug2
;@debug1:
;        mov.l r3,@r1
;@debug2:
;        add r4,r1
;        shlr r0
;        dt r5
;        bf @debugloop


;        imov.l $5F80108,r1
;        mov #2,r0
;        mov.w r0,@r1

;        bsr waitvdp1
;        nop

;        imov.l $5F80108,r1
;        mov #$18,r0
;        mov.w r0,@r1

        bsr waitv
        nop
        bsr setvdp1field
        nop


        ; if not moving or turning, check autoplay script
        mov.w @($0E,r14),r0     ; rotation?
        cmp/eq #0,r0
        bf a164
        mov.l @($30,r14),r0     ; movement?
        cmp/eq #0,r0
        bf a164
        mov.l @($3C,r14),r0     ; script timer?
        cmp/eq #0,r0
        bt a165
        add #-1,r0
;        mov.l r0,@($3C,r14)
        bra a166
        nop
a165:
        mov.l @($38,r14),r1
        mov.w @r1+,r0
        mov.l r1,@($38,r14)
        cmp/pz r0
        bt a166
        shll8 r0
        mov.l r0,@($28,r14)
        not r0,r0
        mov.l r0,@($20,r14)
        bra a164
        nop
a166:
        mov.l r0,@($3C,r14)
a164:



        ; are we  moving?
        mov.l @($30,r14),r0
        cmp/eq #0,r0
        bf a144
        ; check left/right input for changing direction
        imov.l $3FFF,r1
        mov #0,r2
        mov.w @($0C,r14),r0
        and r0,r1
        cmp/eq r1,r2
        bf a141
        mov.l @($20,r14),r0
        shlr8 r0
        tst #4,r0
        bf a133
        mov #-64,r2      ; left
a133:
        tst #8,r0
        bf a134
        mov #64,r2      ; right
a134:
        mov r2,r0
        mov.w r0,@($0E,r14)
        cmp/eq #0,r0
        bf a141
        ; check up/down input for movement
        mov #3,r3
        mov.l @($20,r14),r2
        shlr8 r2
        and r3,r2
        cmp/eq r3,r2
        bt a145
        ; which direction are we facing?
        mov.w @($0C,r14),r0
        cmp/pz r0        
        bf a142
        xor r3,r2
a142:
        imov.l $4000,r1
        tst r1,r0
        bt a143
        xor r3,r2
        shll2 r2
a143:
        mov.l r2,@($30,r14)
        bra a144
        nop
a141:
        ; add angular velocity to axz
        mov.w @($0E,r14),r0
        mov r0,r1
        mov.w @($0C,r14),r0
        add r1,r0
        add r1,r0
        add r1,r0
        add r1,r0
        mov.w r0,@($0C,r14)
        bra a145
        nop

a144:
        ; movement
        mov.l @($30,r14),r1
        mov.l @($2C,r14),r4
        mov #2,r2
        cmp/gt r2,r1
        bt a146
        ; up/down
        swap.w r4,r4
        mov.w @($14,r14),r0
        tst r2,r1
        bt a147
        bsr mvadd
        nop
        bra a148
        nop
a147:
        bsr mvsub
        nop
a148:
        swap.w r4,r4
        mov.w r0,@($14,r14)
        bra a149
        nop
a146:
        ; left/right
        mov.w @($16,r14),r0
        shll r2
        tst r2,r1
        bt a150
        bsr mvsub
        nop
        bra a151
        nop
a150:
        bsr mvadd
        nop
a151:
        mov.w r0,@($16,r14)
a149:
        mov.l r4,@($2C,r14)
        mov.l @($14,r14),r0
        cmp/eq #0,r0
        bf a145
        mov.l r0,@($30,r14)
a145:


        ; check for signal to put up text
        mov.l @($28,r14),r0
        shlr16 r0
        tst #1,r0
        bt a156
        mov.l @($34,r14),r5
        imov.l $5E5B0A0,r6
        bsr drawstring
        nop        
        mov.l r5,@($34,r14)
a156:

        ; check for signal to take down text
        mov.l @($28,r14),r0
        shlr16 r0
        tst #2,r0
        bt a161
        imov.l $5E5B000,r6
        imov.l $5E5F000,r7
        mov #0,r0
a162:
        mov.w r0,@r6
        add #2,r6
        cmp/eq r6,r7
        bf a162
a161:


        ; debugging weriuolhgwie4rugfnlrgf boobs
;        imov.l $5E44200,r2
;        mov.l @($2C,r14),r5

;        mov #7,r1
;        and r5,r1
;        add #27,r1
;        bsr drawtext
;        nop
;        mov #0,r1
;        bsr drawtext
;        shlr16 r5
;        mov #7,r1
;        and r5,r1
;        add #27,r1
;        bsr drawtext
;        nop

        mov.l @(jmpaddr,PC),r1
        jmp @r1
        nop

        idata

jmpaddr:
        dc.l buildtunnel
stadr:
        dc.l sintabl

mvsub:
        add #-4,r0
        and #252,r0
        cmp/eq #124,r0
        bf a152
        add #-1,r4
a152:
        exts.b r0,r0
        rts
        nop

mvadd:
        add #4,r0
        and #252,r0
        imov.l #128,r1
        cmp/eq r1,r0
        bf a153
        add #1,r4
a153:
        exts.b r0,r0
        rts
        nop

waitv:
        mov.l @($1C,r14),r1
@wloop:
        mov.l @($1C,r14),r2
        cmp/eq r1,r2
        bt @wloop
        rts
        nop

waitvdp1:
        mov.l @($1C,r14),r2
        imov.l $25D00010,r1
@wloop:
        mov.l @($1C,r14),r3     ; check for VB also
        cmp/eq r3,r2            ; so we dont get stuck in an infinite loop
        bf @wexit               ; if VDP was too slow
        mov.w @r1,r0
        tst #2,r0
        bt @wloop
@wexit:
        rts
        nop

setvdp1field:
        imov.l $25F80004,r1
        imov.l $25D00002,r2
        mov.w @r1,r0
        not r0,r0
        and #2,r0
        shll r0
        or #8,r0
        mov.w r0,@r2
        rts
        nop

drawstring:
        ; display a string from address in r5
        ; positioned at address in r6

        sts.l PR,@-r15
        mov r6,r7
a157:
;        bra a157
 ;       nop
        mov.b @r5+,r0
        and #255,r0
        imov.l #253,r1
        cmp/eq r0,r1            ; check for end of line
        bf a158
        imov.l $2000,r1
        add r1,r6
        bra a157
        mov r6,r7
a158:
        add #1,r1
        cmp/eq r0,r1            ; check for end of string
        bt a159
        add #-32,r0
        cmp/eq #0,r0
        bt a160
        add #-32,r0
        mov #31,r1
        cmp/gt r1,r0
        bf a160
        add #4,r0
a160:
        mov r0,r1
        bsr drawtext
        mov r7,r2

        bra a157
        add #16,r7

a159:
        lds.l @r15+,PR
        rts
        nop

drawtext:
        ; draw the character in r1, to the address in r2
        ; (0=blank, 1="A", 27="0", etc.)
        mova @(binfont,PC),r0
        mov r0,r13
        shll2 r1
        shll r1
        add r1,r13
        mov #-1,r3
        imov.l $3F0,r4
        mov #8,r12
a126:
        mov.b @r13+,r0
        mov #8,r11
a127:
        mov.w r11,@r2
        tst #128,r0
        bt a128
        mov.w r3,@r2
a128:
        add #2,r2
        shll r0
        dt r11
        bf a127
        add r4,r2
        dt r12
        bf a126
        rts
        nop

        idata

binfont:
        incbin "font.blk"

        align.l
sintabl:
;        incbin "sintabl2.bin"
        incbin "sintabl4.bin"

        align.l
thebmp:
        incbin "mask2.sbc"

;        align.l
;thebmp2:
;        incbin "backdrop.sbc"

        align.l
atexture:
; $8000
        incbin "walltex2.bin"
; $8140
        incbin "walltex3.bin"
; $8280
;        incbin "flortex1.bin"
; $8340
;        incbin "flortex2.bin"

        align.l
player:
        incbin "68play.bin"
samples:
        incbin "samples.bin"

        align.l
        ; bits 3-0 R L D U
dungeon:
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$000,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$000,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$005,$003,$003,$003,$009,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$00C,$000,$000,$000,$00C,$000,0,0,0
        dc.w 0,0,0,$007,$003,$001,$003,$00A,$000,$000,$000,$00C,$000,0,0,0
        dc.w 0,0,0,$000,$000,$00C,$000,$000,$000,$000,$000,$00C,$000,0,0,0
        dc.w 0,0,0,$000,$000,$00C,$000,$000,$000,$000,$000,$00C,$000,0,0,0
        dc.w 0,0,0,$000,$000,$00E,$000,$000,$007,$003,$003,$008,$000,0,0,0

        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$00C,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$00C,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$00C,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$00E,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$000,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$000,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$000,$000,0,0,0
        dc.w 0,0,0,$000,$000,$000,$000,$000,$000,$000,$000,$000,$000,0,0,0

searchlist:
        dc.b 0,-3,0,3,-3,0,3,0
        dc.b 1,-2,1,2,-1,-2,-1,2
        dc.b 2,-1,2,1,-2,-1,-2,1
        dc.b 0,-2,0,2,-2,0,2,0
        dc.b 1,1,1,-1,-1,1,-1,-1
        dc.b 1,0,-1,0,0,-1,0,1
        dc.b 0,0

backscreen:
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084
        dc.w $1084,$1084,$1084,$1084,$1084,$1084,$1084,$1084

        dc.w $0C63,$0C63,$0C63,$0C63,$0C63,$0C63,$0C63,$0C63
        dc.w $0C63,$0C63,$0C63,$0C63,$0C63,$0C63,$0C63,$0C63
        dc.w $0842,$0842,$0842,$0842,$0842,$0842,$0842,$0842
        dc.w $0421,$0421,$0421,$0421,$0421,$0421,$0421,$0421

        dc.w $0000,$0000,$0000,$0000,$0000,$0000,$0000,$0000
        dc.w $0000,$0000,$0000,$0000,$0000,$0000,$0000,$0000
        dc.w $0000,$0000,$0000,$0000,$0000,$0000,$0000,$0000
        dc.w $0000,$0000,$0000,$0000,$0000,$0000,$0000,$0000

        dc.w $0024,$0024,$0024,$0024,$0024,$0024,$0024,$0024
        dc.w $0046,$0046,$0046,$0046,$0046,$0046,$0046,$0046
        dc.w $0068,$0068,$0068,$0068,$0068,$0068,$0068,$0068
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A

        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A
        dc.w $008A,$008A,$008A,$008A,$008A,$008A,$008A,$008A

sometext:
        dc.b "Welcome to THE DUNGEON",124,253
        dc.b "It is a curious place",254

        dc.b "where every wall has",253
        dc.b "the same texture",254

        dc.b "and rendering floors or",253
        dc.b "ceilings is right out",124,254

        dc.b "Here I have made a",253
        dc.b "discovery",126,254

        dc.b "and it is that RISC CPUs",253
        dc.b "are not fun in assembly",124,254

        dc.b "presented at ",141,"party",253
        dc.b "by DamageX",124,124,124," THE END",254

        dc.b "this demo is OVER",127,253
        dc.b "why are you still here",125,254

        dc.b "THE DUNGEON is actually not",253
        dc.b "that curious of a place",254

        dc.b "if you keep pushing buttons",253
        dc.b "garbage may appear onscreen",254

        dc.b "you have been warned",254

        align.w
apscript:
        dc.w $8001,$8100,250,$8200
        dc.w $8004,$8001,$8001,$8100,250,$8200
        dc.w $8008,$8008,$8001,$8001,$8001,$8001,$8100,250,$8200
        dc.w $8004,$8001,$8001,$8100,250,$8200
        dc.w $8008,$8001,$8001,$8001,$8001,$8100,250,$8200
        dc.w $8008,$8001,$8001,$8001,$8001,$8001,$8001,$8002,$8008
        dc.w $8001,$8001,$8001,$8100
        dc.w $7FFF,$7FFF,$7FFF,$7FFF,$7FFF,$7FFF,$7FFF

